from typing import Any

from pydantic import Field

from dhenara.ai.types.external_api._providers import AIModelAPIProviderEnum, AIModelProviderEnum
from dhenara.ai.types.shared.api import SSEEventType, SSEResponse
from dhenara.ai.types.shared.base import BaseModel

from ._content_item import ChatResponseContentItem, ChatResponseContentItemDelta, UsageCharge


class AIModelCallResponseMetaData(BaseModel):
    streaming: bool = False
    duration_seconds: int | float | None = None
    provider_metadata: dict | None = None


class ChatResponseChoice(BaseModel):
    """A single choice/completion in the chat response"""

    index: int
    finish_reason: Any | None = None
    stop_sequence: Any | None = None
    contents: list[ChatResponseContentItem] | None = None
    metadata: dict = {}

    class Config:
        json_schema_extra = {
            "example": {
                "index": 0,
                "contents": [
                    {
                        "role": "assistant",
                        "text": "Hello! How can I help you today?",
                    }
                ],
            }
        }


class ChatResponseChoiceDelta(BaseModel):
    """A single choice/completion in the chat response"""

    index: int
    finish_reason: Any | None = None
    stop_sequence: Any | None = None
    content_deltas: list[ChatResponseContentItemDelta] | None = None
    metadata: dict = {}


class ChatResponseUsage(BaseModel):
    """Token usage statistics for the chat completion"""

    total_tokens: int
    prompt_tokens: int
    completion_tokens: int

    class Config:
        json_schema_extra = {
            "example": {
                "total_tokens": 100,
                "prompt_tokens": 50,
                "completion_tokens": 50,
            }
        }


class ChatResponse(BaseModel):
    """Complete chat response from an AI model

    Contains the response content, usage statistics, and provider-specific metadata
    """

    model: str
    provider: AIModelProviderEnum
    api_provider: AIModelAPIProviderEnum | None = None
    usage: ChatResponseUsage | None = None
    usage_charge: UsageCharge | None = None
    choices: list[ChatResponseChoice] = []
    metadata: AIModelCallResponseMetaData | dict = {}

    def get_visible_fields(self) -> dict:
        return self.model_dump(exclude=["choices"])


class ChatResponseChunk(BaseModel):
    """Chat response Chunk from an AI model

    Contains the response content, usage statistics, and provider-specific metadata
    """

    model: str
    provider: AIModelProviderEnum
    api_provider: AIModelAPIProviderEnum | None = None
    usage: ChatResponseUsage | None = None
    usage_charge: UsageCharge | None = None
    choice_deltas: list[ChatResponseChoiceDelta] = []
    metadata: AIModelCallResponseMetaData | dict = {}

    done: bool = Field(
        default=False,
        description="Indicates if this is the final chunk",
    )

    def get_visible_fields(self) -> dict:
        return self.model_dump(exclude=["choices"])


class StreamingChatResponse(SSEResponse[ChatResponseChunk]):
    """Specialized SSE response for chat streaming"""

    event: SSEEventType = SSEEventType.TOKEN_STREAM
    data: ChatResponseChunk



from pydantic import Field

from dhenara.ai.types.external_api._providers import AIModelProviderEnum
from dhenara.ai.types.shared.base import BaseModel

from ._chat import AIModelCallResponseMetaData
from ._content_item import ImageResponseContentItem, UsageCharge


class ImageResponseChoice(BaseModel):
    """A single image generation choice/result"""

    index: int
    contents: list[ImageResponseContentItem] = []

    class Config:
        json_schema_extra = {
            "example": {
                "index": 0,
                "content": {
                    "content_format": "url",
                    "content_url": "https://api.example.com/images/123.jpg",
                },
            }
        }


class ImageResponseUsage(BaseModel):
    """Usage information for image generation.
    Note that, for images, no usage data is received, so this class holds params required for usage/cost calculation"""

    number_of_images: int = Field(
        ...,
        description="Number of Images generated",
    )
    model: str = Field(
        default_factory=dict,
        description="Model Name",
    )
    options: dict = Field(
        default_factory=dict,
        description="Options send to API",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "model": "dall-e-3",
                "options": {
                    "size": "1024x1024",
                    "quality": "standard",
                },
            }
        }


class ImageResponse(BaseModel):
    """Complete response from an AI image generation model

    Contains the generated images, usage information, and provider-specific metadata
    """

    model: str
    provider: AIModelProviderEnum
    usage: ImageResponseUsage | None
    usage_charge: UsageCharge | None
    choices: list[ImageResponseChoice]
    metadata: AIModelCallResponseMetaData | dict = {}



import base64
from typing import Union

from pydantic import Field

from dhenara.ai.types.shared.base import BaseEnum, BaseModel


class ImageContentFormat(BaseEnum):
    """Enum representing different formats of image content"""

    URL = "url"
    BASE64 = "base64"
    BYTES = "bytes"
    UNKNOWN = "unknown"


class BaseResponseContentItem(BaseModel):
    """Base content item for AI model responses

    Contains common metadata fields used across different types of AI responses

    Attributes:
        metadata: System-generated metadata from API response
        storage_metadata: Storage-related metadata (e.g., cloud storage information)
        custom_metadata: User-defined additional metadata
    """

    index: int = Field(
        default=0,
        description="Content item index",
    )
    metadata: dict = Field(
        default_factory=dict,
        description="System-generated metadata from API response processing",
    )
    storage_metadata: dict = Field(
        default_factory=dict,
        description=(
            "User-defined storage-related metadata such as cloud storage details, paths, or references. "
            "Will be empty on output from `dhenara-ai` package."
        ),
    )
    custom_metadata: dict = Field(
        default_factory=dict,
        description=(
            "User-defined additional metadata for custom processing or tracking."
            "Will be empty on output from `dhenara-ai` package"
        ),
    )


class ChatResponseContentItemType(BaseEnum):
    """Enum representing different types of content items in chat responses"""

    TEXT = "text"
    REASONING = "reasoning"
    GENERIC = "generic"
    TOOL_CALL = "tool_call"


class BaseChatResponseContentItem(BaseResponseContentItem):
    type: ChatResponseContentItemType = Field(
        ...,
        description="Type of the content item",
    )
    role: str | None = Field(
        default=None,
        description="Role of the message sender in the chat context",
    )


class ChatResponseTextContentItem(BaseChatResponseContentItem):
    """Content item specific to chat responses

    Contains the role, text content, and optional function calls for chat interactions

    Attributes:
        role: The role of the message sender (system, user, assistant, or function)
        text: The actual text content of the message
        function_call: Optional function call details if the message involves function calling
    """

    type: ChatResponseContentItemType = ChatResponseContentItemType.TEXT

    text: str | None = Field(
        None,
        description="Plain text content of the message for chat interaction (without reasoning)",
    )

    def get_text(self) -> str:
        return self.text


class ChatResponseReasoningContentItem(BaseChatResponseContentItem):
    type: ChatResponseContentItemType = ChatResponseContentItemType.REASONING

    thinking_text: str | None = Field(
        None,
        description="Thinking text content, for reasoning mode",
    )

    def get_text(self) -> str:
        return self.thinking_text


class ChatResponseToolCallContentItem(BaseChatResponseContentItem):
    type: ChatResponseContentItemType = ChatResponseContentItemType.TOOL_CALL
    # Use metadata to store the content
    # TODO

    def get_text(self) -> str:
        return str(self.metadata)


class ChatResponseGenericContentItem(BaseChatResponseContentItem):
    type: ChatResponseContentItemType = ChatResponseContentItemType.GENERIC
    # Use metadata to store the content

    def get_text(self) -> str:
        return str(self.metadata)


ChatResponseContentItem = Union[  # noqa: UP007
    ChatResponseTextContentItem,
    ChatResponseReasoningContentItem,
    ChatResponseToolCallContentItem,
    ChatResponseGenericContentItem,
]


# Deltas for streamin
class BaseChatResponseContentItemDelta(BaseResponseContentItem):
    type: ChatResponseContentItemType = Field(
        ...,
        description="Type of the content item",
        serialization_alias="type",  # Ensures type is serialized correctly
    )
    role: str | None = Field(
        default=None,
        description="Role of the message sender in the chat context",
    )


class ChatResponseTextContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.TEXT

    text_delta: str | None = Field(
        None,
    )

    def get_text_delta(self) -> str:
        return self.text_delta


class ChatResponseReasoningContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.REASONING

    thinking_text_delta: str | None = Field(
        None,
    )

    def get_text_delta(self) -> str:
        return self.thinking_text_delta


class ChatResponseToolCallContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.TOOL_CALL
    # Use metadata to store the content
    # TODO

    def get_text_delta(self) -> str:
        return str(self.metadata)


class ChatResponseGenericContentItemDelta(BaseChatResponseContentItemDelta):
    type: ChatResponseContentItemType = ChatResponseContentItemType.GENERIC
    # Use metadata to store the content

    def get_text_delta(self) -> str:
        return str(self.metadata)


ChatResponseContentItemDelta = Union[  # noqa: UP007
    ChatResponseTextContentItemDelta,
    ChatResponseReasoningContentItemDelta,
    ChatResponseToolCallContentItemDelta,
    ChatResponseGenericContentItemDelta,
]


class ImageResponseContentItem(BaseResponseContentItem):
    """Content item specific to image generation responses

    Contains the generated image data in various formats (bytes, base64, or URL)

    Attributes:
        content_bytes: Raw image bytes
        content_b64_json: Base64 encoded image data
        content_url: URL to the generated image
        format: Image format (e.g., PNG, JPEG)
        size: Image dimensions
    """

    content_format: ImageContentFormat = Field(
        ...,
        description="Response content format",
    )
    content_bytes: bytes | None = Field(
        None,
        description="Raw image content in bytes",
    )
    content_b64_json: str | None = Field(
        None,
        description="Base64 encoded image content",
        min_length=1,
    )
    content_url: str | None = Field(
        None,
        description="URL to access the generated image",
        pattern=r"^https?://.*$",
    )

    def validate_content(self) -> bool:
        """Validates that at least one content field is populated

        Returns:
            bool: True if at least one content field has data
        """
        return any(
            [
                self.content_bytes is not None,
                self.content_b64_json is not None,
                self.content_url is not None,
            ]
        )

    def get_content_as_bytes(self) -> bytes:
        if self.content_format == ImageContentFormat.BYTES:
            byte_content = self.content_bytes
        elif self.content_format == ImageContentFormat.BASE64:
            byte_content = base64.b64decode(self.content_b64_json)
        else:
            raise ValueError(
                f"get_content_as_bytes: Content format {self.content_format} not supported."
                "Only byte and b64_json is supported now"
            )

        return byte_content


class UsageCharge(BaseModel):
    cost: float = Field(
        ...,
        description="Cost",
    )
    charge: float | None = Field(
        ...,
        description="Charge after considering internal expences and margins."
        " Will be  None if  `cost_multiplier_percentage` is not set in cost data.",
    )



# -----------------------------------------------------------------------------
from dhenara.ai.types.shared.base import BaseEnum


class OpenAiMessageRoleEnum(BaseEnum):
    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"


class GoogleAiMessageRoleEnum(BaseEnum):
    USER = "user"
    MODEL = "model"


class AnthropicMessageRoleEnum(BaseEnum):
    USER = "user"
    ASSISTANT = "assistant"
    # NOTE: There is no "system" role for input messages in the Messages API.
    #  To include a system prompt, you can use the top-level system parameter
    # SYSTEM = "system"



import logging
from typing import Any

from google import genai

from dhenara.ai.providers.base import AIModelProviderClientBase
from dhenara.ai.providers.shared import APIProviderSharedFns
from dhenara.ai.types.external_api import (
    AIModelAPIProviderEnum,
    FormattedPrompt,
    GoogleAiMessageRoleEnum,
    GoogleAIPromptMessage,
    SystemInstructions,
)
from dhenara.ai.types.genai import AIModel
from dhenara.ai.types.shared.file import FileFormatEnum, GenericFile

logger = logging.getLogger(__name__)


class GoogleAIClientBase(AIModelProviderClientBase):
    """Base class for all Google AI Clients"""

    prompt_message_class = GoogleAIPromptMessage

    def initialize(self) -> None:
        pass

    def cleanup(self) -> None:
        pass

    def process_instructions(
        self,
        instructions: SystemInstructions,
    ) -> FormattedPrompt | str | None:
        instructions_str = None
        if instructions:
            if isinstance(instructions, list):
                instructions_str = " ".join(instructions)
            else:
                logger.warning(f"process_instructions: instructions should be a list not {type(instructions)}")
                instructions_str = str(instructions)

            # Some models don't support system instructions
            if any(self.model_endpoint.ai_model.model_name.startswith(model) for model in ["gemini-1.0-pro"]):
                instruction_as_prompt = self.get_prompt(
                    model=self.model_endpoint.ai_model,
                    role=GoogleAiMessageRoleEnum.USER,
                    text=instructions_str,
                    file_contents=[],
                )
                return instruction_as_prompt
        return instructions_str

    def _get_client_params(self, api) -> tuple[str, dict]:
        """Common logic for both sync and async clients"""
        if api.provider == AIModelAPIProviderEnum.GOOGLE_AI:
            return "google_ai", {"api_key": api.api_key}
        elif api.provider == AIModelAPIProviderEnum.GOOGLE_VERTEX_AI:
            client_params = APIProviderSharedFns.get_vertex_ai_credentials(api)
            return "vertex_ai", {
                "vertexai": True,
                "credentials": client_params["credentials"],
                "project": client_params["project_id"],
                "location": client_params["location"],
            }
        else:
            error_msg = f"Unsupported API provider {api.provider} for Google AI"
            logger.error(error_msg)
            raise ValueError(error_msg)

    def _setup_client_sync(self) -> genai.Client:
        """Get the appropriate sync Google AI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)
        return genai.Client(**params)

    async def _setup_client_async(self) -> genai.Client:
        """Get the appropriate async Google AI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)
        return genai.Client(**params).aio

    @staticmethod
    def get_prompt(
        model: AIModel,
        role: GoogleAiMessageRoleEnum,
        text: str,
        file_contents: list,
    ) -> dict:
        # Convert text and file_contents into proper Part format
        parts = []

        # Add the text as a Part
        if text:
            parts.append({"text": text})

        # Add file contents if any
        if file_contents:
            parts.extend(file_contents)

        return {"role": role.value, "parts": parts}

    @staticmethod
    def get_prompt_file_contents(
        model: AIModel,
        files: list[GenericFile],
        max_words: int | None,
    ) -> list[dict[str, Any]]:
        contents = []
        for file in files:
            file_format = file.get_file_format()
            if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                text = f"\nFile: {file.get_source_file_name()}  Content: {file.get_processed_file_data(max_words)}"
                pcontent = text
                contents.append(
                    {
                        "text": pcontent,
                    }
                )
            elif file_format in [FileFormatEnum.IMAGE]:
                mime_type = file.get_mime_type()
                contents.append(
                    {
                        "inline_data": {
                            "data": file.get_processed_file_data_content_only(),  # Bytes type
                            "mime_type": mime_type,
                        },
                    }
                )
            else:
                logger.error(f"get_prompt_file_contents: Unknown file_format {file_format} for file {file.name} ")

        return contents



import logging

from dhenara.ai.providers.anthropic import AnthropicClientBase
from dhenara.ai.providers.google import GoogleAIClientBase
from dhenara.ai.providers.openai import OpenAIClientBase
from dhenara.ai.types.external_api import (
    AIModelProviderEnum,
    AnthropicMessageRoleEnum,
    FormattedPrompt,
    GoogleAiMessageRoleEnum,
    OpenAiMessageRoleEnum,
)
from dhenara.ai.types.genai import AIModel, ChatResponse, ChatResponseChoice, ImageResponse
from dhenara.ai.types.shared.file import GenericFile

logger = logging.getLogger(__name__)


class PromptFormatter:
    """Handles conversion of content into model-specific prompt formats"""

    PROVIDER_CONFIG = {
        AIModelProviderEnum.OPEN_AI: {
            "user_role": OpenAiMessageRoleEnum.USER,
            "assistant_role": OpenAiMessageRoleEnum.ASSISTANT,
            "formatter": OpenAIClientBase,
        },
        AIModelProviderEnum.GOOGLE_AI: {
            "user_role": GoogleAiMessageRoleEnum.USER,
            "assistant_role": GoogleAiMessageRoleEnum.MODEL,
            "formatter": GoogleAIClientBase,
        },
        AIModelProviderEnum.ANTHROPIC: {
            "user_role": AnthropicMessageRoleEnum.USER,
            "assistant_role": AnthropicMessageRoleEnum.ASSISTANT,
            "formatter": AnthropicClientBase,
        },
        AIModelProviderEnum.DEEPSEEK: {
            "user_role": OpenAiMessageRoleEnum.USER,  # Note OpenAI roles
            "assistant_role": OpenAiMessageRoleEnum.ASSISTANT,
            "formatter": OpenAIClientBase,
        },
    }

    @staticmethod
    def _truncate_text_by_words(text: str, max_words: int | None) -> str:
        """Helper method to truncate text by word count"""
        if not max_words or not text:
            return text
        words = text.split()
        return " ".join(words[:max_words])

    @staticmethod
    def format_conversion_node_as_prompts(
        model: AIModel,
        user_query: str | None = None,
        attached_files: list[GenericFile] | None = None,
        previous_response: ChatResponse | ImageResponse | None = None,
        max_words_query: int | None = None,
        max_words_files: int | None = None,
        max_words_response: int | None = None,
        response_before_query: bool = False,
        concat_previous_response_content_items: bool = True,
    ) -> list[FormattedPrompt]:
        """
        Formats conversation elements into model-specific prompts.

        Args:
            model: The AI model being used
            user_query: The user's question or input
            attached_files: list of files to be included in the prompt
            previous_response: Previous model response to include in conversation
            max_words_query: Maximum words for user query
            max_words_files: Maximum words for file content
            max_words_response: Maximum words for previous response
            response_before_query: Whether to place response before query in prompt

        Returns:
            list of formatted prompt messages
        """
        # Extract and truncate previous response if available
        response_messages = []
        _previous_response_text = None
        config = PromptFormatter.PROVIDER_CONFIG[model.provider]
        formatter = config["formatter"]

        if attached_files is None:
            attached_files = []

        if (attached_files and not isinstance(attached_files, list)) or not all(
            isinstance(f, GenericFile) for f in attached_files
        ):
            raise ValueError(f"Invalid type {type(attached_files)} for attached files. Should be list of GenericFile")

        if previous_response:
            try:
                if isinstance(previous_response, dict):
                    _choices = previous_response.get("choices", [])
                    if _choices:
                        choices = [ChatResponseChoice(**choice) for choice in _choices]
                elif isinstance(previous_response, (ChatResponse, ImageResponse)):
                    choices = previous_response.choices
                else:
                    raise ValueError(
                        f"format_as_prompts: previous_response type {type(previous_response)} not supported"
                    )

                _previous_content_items = [content_item for choice in choices for content_item in choice.contents]

                # NOTE: Output files generated in previous response are not handled.
                # They need to be passed in as attached_files to be included in the prompt.
                # TODO: Add docs
                if concat_previous_response_content_items:
                    response_text = None
                    _previous_response_text = " ".join(
                        [
                            f"type:{content_item.type}, content: {content_item.get_text()}"
                            for content_item in _previous_content_items
                        ]
                    )

                    if _previous_response_text:
                        response_text = PromptFormatter._truncate_text_by_words(
                            _previous_response_text, max_words_response
                        )
                    # Format previous response if present
                    if response_text:
                        response_message = formatter.get_prompt(
                            model=model,
                            role=config["assistant_role"],
                            text=response_text,
                            file_contents=[],
                        )
                        response_messages = [response_message]

                else:
                    # TODO_FUTURE: max_words_response has no effect here
                    response_messages = [
                        formatter.get_prompt(
                            model=model,
                            role=config["assistant_role"],
                            text=content_item.get_text(),
                            file_contents=[],
                        )
                        for content_item in _previous_content_items
                    ]

            except Exception as e:
                logger.error(f"Failed to extract previous response text: {e}")

        # Truncate user query if needed
        query_text = PromptFormatter._truncate_text_by_words(user_query, max_words_query) if user_query else None

        # Initialize prompt components
        file_contents = []
        query_message = None

        # Validate provider
        if model.provider not in PromptFormatter.PROVIDER_CONFIG:
            raise ValueError(f"Unsupported model provider: {model.provider}")

        # Format query if present
        if query_text:
            if attached_files:
                file_contents = formatter.get_prompt_file_contents(
                    model=model,
                    files=attached_files,
                    max_words=max_words_files,
                )
            query_message = formatter.get_prompt(
                model=model,
                role=config["user_role"],
                text=query_text,
                file_contents=file_contents,
            )

        # Arrange messages in requested order
        messages = []
        if response_before_query:
            if response_messages:
                messages.extend(response_messages)
            if query_message:
                messages.append(query_message)
        else:
            if query_message:
                messages.append(query_message)
            if response_messages:
                messages.extend(response_messages)

        return messages



import logging
from typing import Any

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI

from dhenara.ai.providers.base import AIModelProviderClientBase
from dhenara.ai.types.external_api import (
    AIModelAPIProviderEnum,
    AIModelFunctionalTypeEnum,
    FormattedPrompt,
    OpenAiMessageRoleEnum,
    OpenAIPromptMessage,
    SystemInstructions,
)
from dhenara.ai.types.genai import AIModel
from dhenara.ai.types.shared.file import FileFormatEnum, GenericFile

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class OpenAIClientBase(AIModelProviderClientBase):
    """Base class for all OpenAI Clients"""

    prompt_message_class = OpenAIPromptMessage

    def initialize(self) -> None:
        pass

    def cleanup(self) -> None:
        pass

    def process_instructions(
        self,
        instructions: SystemInstructions,
    ) -> FormattedPrompt | str | None:
        instructions_str = None
        if instructions:
            if isinstance(instructions, list):
                instructions_str = " ".join(instructions)
            else:
                logger.warning(f"get_model_response: instructions should be a list not {type(instructions)}")
                instructions_str = str(instructions)

            # Beta models won't support system role
            system_role = (
                OpenAiMessageRoleEnum.USER if self.model_endpoint.ai_model.beta else OpenAiMessageRoleEnum.SYSTEM
            )
            instruction_as_prompt = self.get_prompt(
                model=self.model_endpoint.ai_model,
                role=system_role,
                text=instructions_str,
                file_contents=[],
            )
            return instruction_as_prompt
        return instructions_str

    def _get_client_params(self, api) -> tuple[str, dict]:
        """Common logic for both sync and async clients"""
        if api.provider == AIModelAPIProviderEnum.OPEN_AI:
            return "openai", {"api_key": api.api_key}

        elif api.provider == AIModelAPIProviderEnum.MICROSOFT_OPENAI:
            client_params = api.get_provider_credentials()
            return "azure_openai", {
                "api_key": client_params["api_key"],
                "azure_endpoint": client_params["azure_endpoint"],
                "api_version": client_params["api_version"],
            }

        elif api.provider == AIModelAPIProviderEnum.MICROSOFT_AZURE_AI:
            client_params = api.get_provider_credentials()
            return "azure_ai", {
                "endpoint": client_params["azure_endpoint"],
                "credential": client_params["api_key"],
            }

        error_msg = f"Unsupported API provider {api.provider} for OpenAI functions"
        logger.error(error_msg)
        raise ValueError(error_msg)

    def _setup_client_sync(self) -> OpenAI | AzureOpenAI:
        """Get the appropriate sync OpenAI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)

        if client_type == "openai":
            return OpenAI(**params)
        elif client_type == "azure_openai":
            return AzureOpenAI(**params)
        else:  # azure_ai
            from azure.ai.inference import ChatCompletionsClient
            from azure.core.credentials import AzureKeyCredential

            return ChatCompletionsClient(
                endpoint=params["endpoint"],
                credential=AzureKeyCredential(key=params["credential"]),
            )

    async def _setup_client_async(self) -> AsyncOpenAI | AsyncAzureOpenAI:
        """Get the appropriate async OpenAI client"""
        api = self.model_endpoint.api
        client_type, params = self._get_client_params(api)

        if client_type == "openai":
            return AsyncOpenAI(**params)
        elif client_type == "azure_openai":
            return AsyncAzureOpenAI(**params)
        else:  # azure_ai
            from azure.ai.inference.aio import ChatCompletionsClient
            from azure.core.credentials import AzureKeyCredential

            return ChatCompletionsClient(
                endpoint=params["endpoint"],
                credential=AzureKeyCredential(key=params["credential"]),
            )

    # -------------------------------------------------------------------------
    # Static methods
    # -------------------------------------------------------------------------
    @staticmethod
    def get_prompt(
        model: AIModel,
        role: OpenAiMessageRoleEnum,
        text: str,
        file_contents: list[dict[str, Any]],
    ) -> dict[str, Any]:
        if model.functional_type == AIModelFunctionalTypeEnum.IMAGE_GENERATION:
            return OpenAIClientBase.get_prompt_image_model(
                model=model,
                role=role,
                text=text,
                file_contents=file_contents,
            )

        if file_contents:
            content = [
                {
                    "type": "text",
                    "text": text,
                },
                *file_contents,
            ]
        else:
            content = text

        return {"role": role.value, "content": content}

    @staticmethod
    def get_prompt_image_model(
        model: AIModel | None,
        role: OpenAiMessageRoleEnum | None,
        text: str,
        file_contents: list[dict[str, Any]],
    ) -> str:
        if file_contents:
            _file_content = " ".join(file_contents)
            content = text + " " + _file_content
        else:
            content = text

        return content

    # -------------------------------------------------------------------------
    @staticmethod
    def get_prompt_file_contents(
        model: AIModel,
        files: list[GenericFile],
        max_words: int | None,
    ) -> list[dict[str, Any]]:
        if model.functional_type == AIModelFunctionalTypeEnum.IMAGE_GENERATION:
            return OpenAIClientBase.get_prompt_file_contents_image_model(
                model=model,
                files=files,
                max_words=max_words,
            )

        # Eg:
        #        {"type": "text", "text": "What's in this image?"},
        #        {
        #            "type": "image_url",
        #            "image_url": {
        #                "url": "https://upload.wikimedia.org/..boardwalk.jpg",
        #                "url":  f"data:image/jpeg;base64,{base64_image}"
        #            }
        #        },
        contents = []
        for file in files:
            file_format = file.get_file_format()
            try:
                if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                    text = f"\nFile: {file.get_source_file_name()}  Content: {file.get_processed_file_data(max_words)}"
                    pcontent = {
                        "type": "text",
                        "text": text,
                    }
                    contents.append(pcontent)
                elif file_format in [FileFormatEnum.IMAGE]:
                    mime_type = file.get_mime_type()
                    if mime_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
                        raise ValueError(f"Unsupported media type: {mime_type} for file {file.name}")

                    pcontent = {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:{mime_type};base64,{file.get_processed_file_data_content_only()}",
                            # "url": f"data:image/jpeg;base64,{file.signed_url}",
                        },
                    }
                    contents.append(pcontent)
                else:
                    logger.error(f"get_prompt_file_contents: Unknown file_format {file_format} for file {file.name} ")
            except Exception as e:
                logger.error(f"Error processing file {file.name}: {e}")

        return contents

    # -------------------------------------------------------------------------
    @staticmethod
    def get_prompt_file_contents_image_model(
        model: AIModel,
        files: list[GenericFile],
        max_words: int | None,
    ) -> str:
        contents: list[dict[str, Any]] = []
        for file in files:
            file_format = file.get_file_format()
            try:
                if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                    text = (
                        f"\nFile: {file.get_source_file_name()}  "
                        f"Content: {file.get_processed_file_data(max_words=max_words)}"
                    )
                    contents.append(text)
                elif file_format == FileFormatEnum.IMAGE:
                    mime_type = file.get_mime_type()
                    if mime_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
                        raise ValueError(f"Unsupported media type: {mime_type} for file {file.name}")

                    pcontent = f"data:{mime_type};base64,{file.get_processed_file_data_content_only()}"
                    contents.append(pcontent)
                else:
                    logger.error(f"get_prompt_file_contents: Unknown file_format {file_format} for file {file.name}")
            except Exception as e:
                logger.error(f"Error processing file {file.name}: {e}")

        return " ".join(contents)



# ruff: noqa: F401
# Ai model APIs
from ._ai_model_response._content_item import *
from ._ai_model_response._chat import *
from ._ai_model_response._image import *
from ._ai_model_response._model_call import *




from pydantic import Field

from dhenara.ai.types.genai.dhenara import ChatResponse, ImageResponse
from dhenara.ai.types.shared.base import BaseModel
from dhenara.ai.types.shared.file import GenericFile


class ConversationNode(BaseModel):
    """Represents a single turn in a conversation."""
    user_query: str
    attached_files: list[GenericFile] = Field(default_factory=list)
    response: ChatResponse | ImageResponse | None = None
    timestamp: str | None = None



import logging
from typing import Annotated, Any, Literal, Union

from pydantic import Field, field_validator

from dhenara.ai.types.external_api import (
    AnthropicMessageRoleEnum,
    GoogleAiMessageRoleEnum,
    OpenAiMessageRoleEnum,
)
from dhenara.ai.types.shared.base import BaseModel

logger = logging.getLogger(__name__)


# OpenAI Specific Models


class OpenAITextContent(BaseModel):
    type: Literal["text"] = "text"
    text: str


class OpenAIImageUrlContent(BaseModel):
    url: str


class OpenAIImageContent(BaseModel):
    type: Literal["image_url"] = "image_url"
    image_url: OpenAIImageUrlContent


# Discriminated union for content parts
ContentPart = Annotated[
    OpenAITextContent | OpenAIImageContent,
    Field(discriminator="type"),
]


class OpenAIPromptMessage(BaseModel):
    role: OpenAiMessageRoleEnum
    content: str | list[ContentPart]
    function_call: dict[str, Any] | None = None

    @field_validator("content", mode="before")
    @classmethod
    def validate_content(cls, v):
        # Handle string case
        if isinstance(v, str):
            return v

        # Handle list case with automatic discrimination
        if isinstance(v, list):
            return [{"type": item.get("type"), **item} for item in v]

        raise ValueError("Content must be string or list of content parts")

    ## fmt: off
    # def model_dump(self, **kwargs):
    #    # Clean None values and empty lists
    #    data = super().model_dump(**kwargs)

    #    if isinstance(data["content"], list):
    #        data["content"] = [
    #            part for part in data["content"]
    #            if (part["type"] == "text" and part["text"]) or
    #               (part["type"] == "image_url" and part["image_url"]["url"])
    #            ]

    #    if not data["content"] and isinstance(data["content"], list):
    #        data["content"] = ""  # Fallback to empty string

    #    return data


# Anthropic Specific Models


class AnthropicMessageTextContent(BaseModel):
    type: Literal["text"]
    text: str


class AnthropicMessageImageContent(BaseModel):
    type: Literal["image"]
    source: dict  # Or use a more specific type for image source


class AnthropicPromptMessage(BaseModel):
    role: AnthropicMessageRoleEnum
    content: str | list[AnthropicMessageTextContent | AnthropicMessageImageContent]

    @field_validator("content")
    @classmethod
    def validate_content(cls, v):
        if isinstance(v, str):
            return v

        for item in v:
            if isinstance(item, dict):
                if item["type"] == "text":
                    if "source" in item:
                        item.pop("source")  # Remove source field for text content
                elif item["type"] == "image" and "source" not in item:
                    raise ValueError("source is required for image type")
        return v


# Google AI Specific Models


class GoogleAIInlineData(BaseModel):
    """Model for inline data like images"""

    data: str
    mime_type: str


class GoogleAIPart(BaseModel):
    """Base model for different types of content parts"""

    text: str | None = None
    inline_data: GoogleAIInlineData | None = None

    @field_validator("inline_data", mode="before")
    @classmethod
    def validate_inline_data(cls, v: Any) -> GoogleAIInlineData | None:
        if v is None:
            return None
        if isinstance(v, dict):
            return GoogleAIInlineData(**v)
        return v

    # def model_dump(self) -> dict[str, Any]:
    #    if self.text is not None:
    #        return {"text": self.text}
    #    if self.inline_data is not None:
    #        return {"inline_data": self.inline_data.model_dump()}
    #    return {}


class GoogleAIPromptMessage(BaseModel):
    """A complete Google AI prompt message"""

    role: GoogleAiMessageRoleEnum
    parts: list[GoogleAIPart]

    @field_validator("parts", mode="before")
    @classmethod
    def validate_parts(cls, v: Any) -> list[GoogleAIPart]:
        if isinstance(v, str):
            return [GoogleAIPart(text=v)]

        if isinstance(v, list):
            validated_parts = []
            for part in v:
                if isinstance(part, str):
                    validated_parts.append(GoogleAIPart(text=part))
                elif isinstance(part, dict):
                    validated_parts.append(GoogleAIPart(**part))
                elif isinstance(part, GoogleAIPart):
                    validated_parts.append(part)
                else:
                    raise ValueError(f"Invalid part type: {type(part)}")
            return validated_parts

        raise ValueError(f"Invalid parts type: {type(v)}")

    # def model_dump(self) -> dict[str, Any]:
    #    return {"role": self.role.value, "parts": [part.model_dump() for part in self.parts]}


# # Configuration Models
# class PromptConfig(BaseModel):
#     model_provider: AIModelProviderEnum
#     max_tokens_query: int | None = Field(default=None, gt=0)
#     max_tokens_files: int | None = Field(default=None, gt=0)
#     max_tokens_response: int | None = Field(default=None, gt=0)
#     response_before_query: bool = False
#
#
# class ConversationNodeContent(BaseModel):
#     user_query: str | None = None
#     attached_files: list[GenericFile] | None = None
#     previous_response: Union[ChatResponse, ImageResponse] | None = None


FormattedPrompt = Union[OpenAIPromptMessage, GoogleAIPromptMessage, AnthropicPromptMessage]  # noqa: UP007
SystemInstructions = list[str]



import logging

from google.genai.types import Content, GenerateContentConfig, GenerateContentResponse, Part, SafetySetting

# Copyright 2024-2025 Dhenara Inc. All rights reserved.
from dhenara.ai.providers.google import GoogleAIClientBase
from dhenara.ai.types.genai import (
    AIModelCallResponse,
    AIModelCallResponseMetaData,
    ChatResponse,
    ChatResponseChoice,
    ChatResponseChoiceDelta,
    ChatResponseContentItem,
    ChatResponseContentItemDelta,
    ChatResponseGenericContentItem,
    ChatResponseGenericContentItemDelta,
    ChatResponseTextContentItem,
    ChatResponseTextContentItemDelta,
    ChatResponseUsage,
    StreamingChatResponse,
)
from dhenara.ai.types.shared.api import SSEErrorResponse

logger = logging.getLogger(__name__)

logger = logging.getLogger(__name__)


models_not_supporting_system_instructions = ["gemini-1.0-pro"]


# -----------------------------------------------------------------------------
class GoogleAIChat(GoogleAIClientBase):
    def get_api_call_params(
        self,
        prompt: dict,
        context: list[dict] | None = None,
        instructions: list[str] | None = None,
    ) -> AIModelCallResponse:
        if not self._client:
            raise RuntimeError("Client not initialized. Use with 'async with' context manager")

        if self._input_validation_pending:
            raise ValueError("inputs must be validated with `self.validate_inputs()` before api calls")

        generate_config_args = self.get_default_generate_config_args()
        generate_config = GenerateContentConfig(**generate_config_args)

        # Process instructions
        instructions_str = self.process_instructions(instructions)
        if isinstance(instructions_str, dict):
            if context:
                context.insert(0, instructions_str)
            else:
                context = [instructions_str]
        elif instructions_str and not any(
            self.model_endpoint.ai_model.model_name.startswith(model) for model in ["gemini-1.0-pro"]
        ):
            generate_config.system_instruction = instructions_str

        history = []
        if context:
            for item in context:
                if isinstance(item, dict):
                    parts = [{"text": p["text"]} for p in item["parts"]] if "parts" in item else []
                    history.append(Content(role=item["role"], parts=parts))
                else:
                    history.append(item)

        return {
            "prompt": prompt,
            "history": history,
            "generate_config": generate_config,
        }

    def do_api_call_sync(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        chat = self._client.chats.create(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            history=api_call_params["history"],
        )
        response = chat.send_message(message=api_call_params["prompt"])
        return response

    async def do_api_call_async(
        self,
        api_call_params: dict,
    ) -> AIModelCallResponse:
        chat = self._client.chats.create(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            history=api_call_params["history"],
        )
        response = await chat.send_message(message=api_call_params["prompt"])
        return response

    def do_streaming_api_call_sync(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        chat = self._client.chats.create(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            history=api_call_params["history"],
        )
        stream = chat.send_message_stream(message=api_call_params["prompt"])
        return stream

    async def do_streaming_api_call_async(
        self,
        api_call_params,
    ) -> AIModelCallResponse:
        chat = self._client.chats.create(
            model=self.model_name_in_api_calls,
            config=api_call_params["generate_config"],
            history=api_call_params["history"],
        )
        stream = await chat.send_message_stream(message=api_call_params["prompt"])
        return stream

    def get_default_generate_config_args(self) -> dict:
        max_output_tokens, max_reasoning_tokens = self.config.get_max_output_tokens(self.model_endpoint.ai_model)
        safety_settings = [
            SafetySetting(
                category="HARM_CATEGORY_DANGEROUS_CONTENT",
                threshold="BLOCK_ONLY_HIGH",
            )
        ]

        config_params = {
            "candidate_count": 1,
            "safety_settings": safety_settings,
        }

        if max_output_tokens:
            config_params["max_output_tokens"] = max_output_tokens

        return config_params

    def parse_stream_chunk(
        self,
        chunk: GenerateContentResponse,
    ) -> StreamingChatResponse | SSEErrorResponse | None:
        """Handle streaming response with progress tracking and final response"""

        processed_chunks = []

        self.streaming_manager.provider_metadata = None

        # Process content
        if chunk.candidates:
            choice_deltas = []
            for candidate_index, candidate in enumerate(chunk.candidates):
                content_deltas = []
                for part_index, part in enumerate(candidate.content.parts):
                    content_deltas.append(
                        self.process_content_item_delta(
                            index=part_index,
                            role=candidate.content.role,
                            delta=part,
                        )
                    )
                choice_deltas.append(
                    ChatResponseChoiceDelta(
                        index=candidate_index,
                        finish_reason=candidate.finish_reason,
                        stop_sequence=None,
                        content_deltas=content_deltas,
                        metadata={"safety_ratings": candidate.safety_ratings, "": candidate.content},  # Choice metadata
                    )
                )

            response_chunk = self.streaming_manager.update(choice_deltas=choice_deltas)
            stream_response = StreamingChatResponse(
                id=None,  # No 'id' from google
                data=response_chunk,
            )

            processed_chunks.append(stream_response)

            # Check if this is the final chunk
            is_done = bool(candidate.finish_reason)

            if is_done:
                usage = self._get_usage_from_provider_response(chunk)
                self.streaming_manager.update_usage(usage)

        return processed_chunks

    def _get_usage_from_provider_response(
        self,
        response: GenerateContentResponse,
    ) -> ChatResponseUsage:
        return ChatResponseUsage(
            total_tokens=response.usage_metadata.total_token_count,
            prompt_tokens=response.usage_metadata.prompt_token_count,
            completion_tokens=response.usage_metadata.candidates_token_count,
        )

    def parse_response(
        self,
        response: GenerateContentResponse,
    ) -> ChatResponse:
        usage, usage_charge = self.get_usage_and_charge(response)
        return ChatResponse(
            model=response.model_version,
            provider=self.model_endpoint.ai_model.provider,
            api_provider=self.model_endpoint.api.provider,
            usage=usage,
            usage_charge=usage_charge,
            choices=[
                ChatResponseChoice(
                    index=choice_index,
                    finish_reason=candidate.finish_reason,
                    stop_sequence=None,
                    contents=[
                        self.process_content_item(
                            index=part_index,
                            role=candidate.content.role,
                            content_item=part,
                        )
                        for part_index, part in enumerate(candidate.content.parts)
                    ],
                    metadata={},  # Choice metadata
                )
                for choice_index, candidate in enumerate(response.candidates)
            ],
            metadata=AIModelCallResponseMetaData(
                streaming=False,
                duration_seconds=0,  # TODO
                provider_metadata={},
            ),
        )

    def process_content_item(
        self,
        index: int,
        role: str,
        content_item: Part,
    ) -> ChatResponseContentItem:
        if isinstance(content_item, Part):
            if content_item.text:
                return ChatResponseTextContentItem(
                    index=index,
                    role=role,
                    text=content_item.text,
                )
            else:
                return ChatResponseGenericContentItem(
                    index=index,
                    role=role,
                    metadata={"part": content_item.model_dump()},
                )
        else:
            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=content_item,
                streaming=False,
            )

    # Streaming
    def process_content_item_delta(
        self,
        index: int,
        role: str,
        delta,
    ) -> ChatResponseContentItemDelta:
        if isinstance(delta, Part):
            if delta.text:
                return ChatResponseTextContentItemDelta(
                    index=index,
                    role=role,
                    text_delta=delta.text,
                )
            else:
                return ChatResponseGenericContentItemDelta(
                    index=index,
                    role=role,
                    metadata={"part": delta.model_dump()},
                )

        else:
            return self.get_unknown_content_type_item(
                index=index,
                role=role,
                unknown_item=delta,
                streaming=True,
            )



import logging
from typing import Any

from anthropic import (
    Anthropic,
    AnthropicBedrock,
    AnthropicVertex,
    AsyncAnthropic,
    AsyncAnthropicBedrock,
    AsyncAnthropicVertex,
)

from dhenara.ai.providers.base import AIModelProviderClientBase
from dhenara.ai.providers.shared import APIProviderSharedFns
from dhenara.ai.types.external_api import (
    AIModelAPIProviderEnum,
    AnthropicMessageRoleEnum,
    AnthropicPromptMessage,
    FormattedPrompt,
    SystemInstructions,
)
from dhenara.ai.types.genai import AIModel
from dhenara.ai.types.shared.file import FileFormatEnum, GenericFile

logger = logging.getLogger(__name__)


class AnthropicClientBase(AIModelProviderClientBase):
    """Base class for all Anthropic Clients"""

    prompt_message_class = AnthropicPromptMessage

    def initialize(self) -> None:
        pass

    def cleanup(self) -> None:
        pass

    def process_instructions(
        self,
        instructions: SystemInstructions,
    ) -> FormattedPrompt | str | None:
        if instructions:
            if isinstance(instructions, list):
                return " ".join(instructions)
            logger.warning(f"process_instructions: instructions should be a list not {type(instructions)}")
            return str(instructions)
        return None

    def _get_client_params(self, api) -> tuple[str, dict]:
        """Common logic for both sync and async clients"""
        if api.provider == AIModelAPIProviderEnum.ANTHROPIC:
            return "anthropic", {"api_key": api.api_key}

        elif api.provider == AIModelAPIProviderEnum.GOOGLE_VERTEX_AI:
            client_params = APIProviderSharedFns.get_vertex_ai_credentials(api)
            return "vertex_ai", {
                "credentials": client_params["credentials"],
                "project_id": client_params["project_id"],
                "region": client_params["location"],
            }

        elif api.provider == AIModelAPIProviderEnum.AMAZON_BEDROCK:
            client_params = api.get_provider_credentials()
            return "bedrock", {
                "aws_access_key": client_params["aws_access_key"],
                "aws_secret_key": client_params["aws_secret_key"],
                "aws_region": client_params.get("aws_region", "us-east-1"),
            }

        error_msg = f"Unsupported API provider {api.provider} for Anthropic"
        logger.error(error_msg)
        raise ValueError(error_msg)

    def _setup_client_sync(self) -> Anthropic | AnthropicBedrock | AnthropicVertex:
        """Get the appropriate sync Anthropic client based on the provider"""
        client_type, params = self._get_client_params(self.model_endpoint.api)

        if client_type == "anthropic":
            return Anthropic(**params)
        elif client_type == "vertex_ai":
            return AnthropicVertex(**params)
        else:  # bedrock
            return AnthropicBedrock(**params)

    async def _setup_client_async(self) -> AsyncAnthropic | AsyncAnthropicBedrock | AsyncAnthropicVertex:
        """Get the appropriate async Anthropic client based on the provider"""
        client_type, params = self._get_client_params(self.model_endpoint.api)

        if client_type == "anthropic":
            return AsyncAnthropic(**params)
        elif client_type == "vertex_ai":
            return AsyncAnthropicVertex(**params)
        else:  # bedrock
            return AsyncAnthropicBedrock(**params)

    @staticmethod
    def get_prompt(
        model: AIModel,
        role: AnthropicMessageRoleEnum,
        text: str,
        file_contents: list[dict[str, Any]],
    ) -> dict[str, Any]:
        if file_contents:
            # Combine text and file contents into a single content array
            content = [
                {
                    "type": "text",
                    "text": text,
                },
                *file_contents,
            ]
        else:
            # For text-only messages, content should be a string
            content = text

        return {"role": role.value, "content": content}

    @staticmethod
    def get_prompt_file_contents(
        model: AIModel,
        files: list[GenericFile],
        max_words: int | None,
    ) -> list[dict[str, Any]]:
        contents: list[dict[str, Any]] = []
        for file in files:
            file_format = file.get_file_format()
            try:
                if file_format in [FileFormatEnum.COMPRESSED, FileFormatEnum.TEXT]:
                    text = (
                        f"\nFile: {file.get_source_file_name()}  "
                        f"Content: {file.get_processed_file_data(max_words=max_words)}"
                    )
                    contents.append(
                        {
                            "type": "text",
                            "text": text,
                        }
                    )
                elif file_format == FileFormatEnum.IMAGE:
                    mime_type = file.get_mime_type()
                    if mime_type not in ["image/jpeg", "image/png", "image/gif", "image/webp"]:
                        raise ValueError(f"Unsupported media type: {mime_type}")

                    contents.append(
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": mime_type,
                                "data": file.get_processed_file_data_content_only(),
                            },
                        }
                    )
                else:
                    logger.error(f"Unknown file_format {file_format} for file {file.name}")
            except Exception as e:
                logger.error(f"Error processing file {file.name}: {e}")

        return contents



from enum import Enum
from typing import Any, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, alias_generators

T = TypeVar("T", bound="BaseModel")
# logger.debug(f"Pydantic version: {pydantic.__version__}")


class BaseEnum(str, Enum):
    """Base Enumeration class."""

    def __str__(self):
        return self.value

    @classmethod
    def values(cls) -> set[str]:
        """Get all values.

        Returns:
            set[str]: Set of all values
        """
        return {member.value for member in cls}


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(
        alias_generator=alias_generators.to_camel,
        populate_by_name=True,
        from_attributes=True,
        protected_namespaces=set(),
        # Enable detailed validation errors:
        validate_assignment=True,
        extra="forbid",
        arbitrary_types_allowed=True,
        # schema etra
        json_schema_extra={"examples": []},
        str_strip_whitespace=False,  # Don't set: Streaming responses will be terrible
        use_enum_values=True,
    )

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        dump_kwargs = {
            "exclude_unset": False,
            "by_alias": False,
            "exclude_none": True,
            "round_trip": False,
        }
        dump_kwargs.update(kwargs)

        return super().model_dump(**dump_kwargs)

    # def copy_with_changes(self: T, **changes) -> T:
    #    """Create a copy with specified changes."""
    #    data = self.model_dump()
    #    data.update(changes)
    #    return self.__class__.model_validate(data)

    # @classmethod
    # def safe_parse(cls: Type[T], data: dict[str, Any]) -> tuple[Optional[T], Optional[ValidationError]]:
    #    """Safely parse data without raising exceptions."""
    #    try:
    #        return cls.model_validate(data), None
    #    except ValidationError as e:
    #        return None, e
